"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const TurndownService = require("turndown");
const LSP = require("vscode-languageserver");
const analyser_1 = require("./analyser");
const Builtins = require("./builtins");
const config = require("./config");
const executables_1 = require("./executables");
const parser_1 = require("./parser");
const ReservedWords = require("./reservedWords");
const types_1 = require("./types");
/**
 * The BashServer glues together the separate components to implement
 * the various parts of the Language Server Protocol.
 */
class BashServer {
    constructor(connection, executables, analyzer) {
        this.documents = new LSP.TextDocuments();
        this.connection = connection;
        this.executables = executables;
        this.analyzer = analyzer;
    }
    /**
     * Initialize the server based on a connection to the client and the protocols
     * initialization parameters.
     */
    static initialize(connection, { rootPath }) {
        return __awaiter(this, void 0, void 0, function* () {
            const parser = yield parser_1.initializeParser();
            return Promise.all([
                executables_1.default.fromPath(process.env.PATH),
                analyser_1.default.fromRoot({ connection, rootPath, parser }),
            ]).then(xs => {
                const executables = xs[0];
                const analyzer = xs[1];
                return new BashServer(connection, executables, analyzer);
            });
        });
    }
    /**
     * Register handlers for the events from the Language Server Protocol that we
     * care about.
     */
    register(connection) {
        // The content of a text document has changed. This event is emitted
        // when the text document first opened or when its content has changed.
        this.documents.listen(this.connection);
        this.documents.onDidChangeContent(change => {
            const { uri } = change.document;
            const diagnostics = this.analyzer.analyze(uri, change.document);
            if (config.getHighlightParsingError()) {
                connection.sendDiagnostics({
                    uri: change.document.uri,
                    diagnostics,
                });
            }
        });
        // Register all the handlers for the LSP events.
        connection.onHover(this.onHover.bind(this));
        connection.onDefinition(this.onDefinition.bind(this));
        connection.onDocumentSymbol(this.onDocumentSymbol.bind(this));
        connection.onWorkspaceSymbol(this.onWorkspaceSymbol.bind(this));
        connection.onDocumentHighlight(this.onDocumentHighlight.bind(this));
        connection.onReferences(this.onReferences.bind(this));
        connection.onCompletion(this.onCompletion.bind(this));
        connection.onCompletionResolve(this.onCompletionResolve.bind(this));
    }
    /**
     * The parts of the Language Server Protocol that we are currently supporting.
     */
    capabilities() {
        return {
            // For now we're using full-sync even though tree-sitter has great support
            // for partial updates.
            textDocumentSync: this.documents.syncKind,
            completionProvider: {
                resolveProvider: true,
            },
            hoverProvider: true,
            documentHighlightProvider: true,
            definitionProvider: true,
            documentSymbolProvider: true,
            workspaceSymbolProvider: true,
            referencesProvider: true,
        };
    }
    getWordAtPoint(params) {
        return this.analyzer.wordAtPoint(params.textDocument.uri, params.position.line, params.position.character);
    }
    logRequest({ request, params, word, }) {
        const wordLog = word ? `"${word}"` : '';
        this.connection.console.log(`${request} ${params.position.line}:${params.position.character} ${wordLog}`);
    }
    onHover(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const word = this.getWordAtPoint(params);
            this.logRequest({ request: 'onHover', params, word });
            const explainshellEndpoint = config.getExplainshellEndpoint();
            if (explainshellEndpoint) {
                this.connection.console.log(`Query ${explainshellEndpoint}`);
                const response = yield this.analyzer.getExplainshellDocumentation({
                    params,
                    endpoint: explainshellEndpoint,
                });
                if (response.status === 'error') {
                    this.connection.console.log(`getExplainshellDocumentation returned: ${JSON.stringify(response, null, 4)}`);
                }
                else {
                    return {
                        contents: {
                            kind: 'markdown',
                            value: new TurndownService().turndown(response.helpHTML),
                        },
                    };
                }
            }
            const getMarkdownHoverItem = (doc) => ({
                // LSP.MarkupContent
                value: ['``` man', doc, '```'].join('\n'),
                // Passed as markdown for syntax highlighting
                kind: 'markdown',
            });
            if (Builtins.isBuiltin(word)) {
                return Builtins.documentation(word).then(doc => ({
                    contents: getMarkdownHoverItem(doc),
                }));
            }
            if (ReservedWords.isReservedWord(word)) {
                return ReservedWords.documentation(word).then(doc => ({
                    contents: getMarkdownHoverItem(doc),
                }));
            }
            if (this.executables.isExecutableOnPATH(word)) {
                return this.executables.documentation(word).then(doc => ({
                    contents: getMarkdownHoverItem(doc),
                }));
            }
            return null;
        });
    }
    onDefinition(params) {
        const word = this.getWordAtPoint(params);
        this.logRequest({ request: 'onDefinition', params, word });
        return this.analyzer.findDefinition(word);
    }
    onDocumentSymbol(params) {
        this.connection.console.log(`onDocumentSymbol`);
        return this.analyzer.findSymbols(params.textDocument.uri);
    }
    onWorkspaceSymbol(params) {
        this.connection.console.log('onWorkspaceSymbol');
        return this.analyzer.search(params.query);
    }
    onDocumentHighlight(params) {
        const word = this.getWordAtPoint(params);
        this.logRequest({ request: 'onDocumentHighlight', params, word });
        return this.analyzer
            .findOccurrences(params.textDocument.uri, word)
            .map(n => ({ range: n.range }));
    }
    onReferences(params) {
        const word = this.getWordAtPoint(params);
        this.logRequest({ request: 'onReferences', params, word });
        return this.analyzer.findReferences(word);
    }
    onCompletion(params) {
        const word = this.getWordAtPoint(params);
        this.logRequest({ request: 'onCompletion', params, word });
        const symbolCompletions = this.analyzer.findSymbolCompletions(params.textDocument.uri);
        // TODO: we could do some caching here...
        const reservedWordsCompletions = ReservedWords.LIST.map(reservedWord => ({
            label: reservedWord,
            kind: LSP.SymbolKind.Interface,
            data: {
                name: reservedWord,
                type: types_1.CompletionItemDataType.ReservedWord,
            },
        }));
        const programCompletions = this.executables.list().map((s) => {
            return {
                label: s,
                kind: LSP.SymbolKind.Function,
                data: {
                    name: s,
                    type: types_1.CompletionItemDataType.Executable,
                },
            };
        });
        const builtinsCompletions = Builtins.LIST.map(builtin => ({
            label: builtin,
            kind: LSP.SymbolKind.Interface,
            data: {
                name: builtin,
                type: types_1.CompletionItemDataType.Builtin,
            },
        }));
        // TODO: we have duplicates here (e.g. echo is both a builtin AND have a man page)
        const allCompletions = [
            ...reservedWordsCompletions,
            ...symbolCompletions,
            ...programCompletions,
            ...builtinsCompletions,
        ];
        if (word) {
            if (word.startsWith('#')) {
                // Inside a comment block
                return [];
            }
            // Filter to only return suffixes of the current word
            return allCompletions.filter(item => item.label.startsWith(word));
        }
        return allCompletions;
    }
    onCompletionResolve(item) {
        return __awaiter(this, void 0, void 0, function* () {
            const { data: { name, type }, } = item;
            this.connection.console.log(`onCompletionResolve name=${name} type=${type}`);
            const getMarkdownCompletionItem = (doc) => (Object.assign(Object.assign({}, item), { 
                // LSP.MarkupContent
                documentation: {
                    value: ['``` man', doc, '```'].join('\n'),
                    // Passed as markdown for syntax highlighting
                    kind: 'markdown',
                } }));
            try {
                if (type === types_1.CompletionItemDataType.Executable) {
                    const doc = yield this.executables.documentation(name);
                    return getMarkdownCompletionItem(doc);
                }
                else if (type === types_1.CompletionItemDataType.Builtin) {
                    const doc = yield Builtins.documentation(name);
                    return getMarkdownCompletionItem(doc);
                }
                else if (type === types_1.CompletionItemDataType.ReservedWord) {
                    const doc = yield ReservedWords.documentation(name);
                    return getMarkdownCompletionItem(doc);
                }
                else {
                    return item;
                }
            }
            catch (error) {
                return item;
            }
        });
    }
}
exports.default = BashServer;
//# sourceMappingURL=server.js.map